#pragma once

#include <array>
#include "d3d9.h"
#include <MinHook.h>
#include "SimpleIni.h"

struct hkIDirect3D9;

static const char* inifilename = "FF13Fix.ini";
#define CONFIG_VERSION 4

class Config
{
public:
	Config(const Config&) = delete;
	const Config& operator=(Config& other) = delete;

	Config();

#define SETTING(_type, _func, _var, _section, _defaultval) \
	private: _type _var; \
	public: const _type& Get##_var() const { return _var; };
#include "Settings.h"
#undef SETTING
};

#define DECLARE_HOOK(type, callconv, name, ...) \
	public: type(callconv* True##name)(__VA_ARGS__) = name; \
	private: static type callconv Hook##name(__VA_ARGS__);

class MainContext
{
	MainContext(const MainContext&) = delete;
	const MainContext& operator=(MainContext& other) = delete;

	DECLARE_HOOK(IDirect3D9*, WINAPI, Direct3DCreate9, UINT SDKVersion);
	DECLARE_HOOK(LONG, WINAPI, SetWindowLongA, HWND hWnd, int nIndex, LONG dwNewLong);
	DECLARE_HOOK(LONG, WINAPI, SetWindowLongW, HWND hWnd, int nIndex, LONG dwNewLong);
	DECLARE_HOOK(HWND, WINAPI, CreateWindowExA, DWORD dwExStyle, LPCSTR lpClassName, LPCSTR lpWindowName,
		DWORD dwStyle, int X, int Y, int nWidth, int nHeight, HWND hWndParent, HMENU hMenu, HINSTANCE hInstance, LPVOID lpParam);
	DECLARE_HOOK(HWND, WINAPI, CreateWindowExW, DWORD dwExStyle, LPCWSTR lpClassName, LPCWSTR lpWindowName,
		DWORD dwStyle, int X, int Y, int nWidth, int nHeight, HWND hWndParent, HMENU hMenu, HINSTANCE hInstance, LPVOID lpParam);

public:
	MainContext();
	~MainContext();

	bool ApplyPresentationParameters(D3DPRESENT_PARAMETERS* pPresentationParameters);
	bool ApplyBehaviorFlagsFix(DWORD* flags);
	HRESULT APIENTRY ApplyVertexBufferFix(IDirect3DDevice9* pIDirect3DDevice9, UINT Length, DWORD Usage, DWORD FVF, D3DPOOL Pool, IDirect3DVertexBuffer9** ppVertexBuffer, HANDLE* pSharedHandle);
	bool BehaviorFlagsToString(DWORD BehaviorFlags, std::string* BehaviorFlagsString);

	bool CheckWindow(HWND hWnd);

	void ApplyWndProc(HWND hWnd);
	void ApplyBorderless(HWND hWnd);

	Config config;

private:
	enum AutoFixes : u32
	{
		NONE = 0,
		RESIDENT_EVIL_4,
		KINGS_BOUNTY_LEGEND,
		FINAL_FANTASY_XIII
	};

	void EnableAutoFix();
	AutoFixes autofix = AutoFixes::NONE;

	bool didOneTimeFixes = false;

	const float MAX_FRAME_RATE_LIMIT = 250000.0F;
	byte* SET_FRAMERATE_INGAME_INSTRUCTION_ADDRESS = (byte*)0x00E8D65F;
	byte* CONTINUOUS_SCAN_INSTRUCTION_ADDRESS = (byte*)0x00820868;
	byte* ENEMY_SCAN_BOX_CODE_ADDRESS = (byte*)0x0094C920;

	float* framePacerTargetPtr = NULL;
	UINT backbufferWidth = 0;
	
	void FixBehaviorFlagConflict(const DWORD flags_in, DWORD* flags_out);
	static const std::map<const AutoFixes, const uint32_t> behaviorflags_fixes;

	static LRESULT CALLBACK WindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
	WNDPROC oldWndProc;

	void FFXIIIOneTimeFixes();
	bool FFXIIINOPIngameFrameRateLimitSetter();
	bool FFXIIISetFrameRateVariables();
	void ChangeMemoryProtectionToReadWriteExecute(void* address, const int size);
	void FixMissingEnemyScan();
	void RemoveContinuousControllerScan();
	bool areAlmostTheSame(float a, float b);
};

extern MainContext context;